function [ masks, extra_info ] = cocoGenGTMask(cocoapi, im_id, ...
                                               seg_idxs, crowd)
%COCOGENGTMASK generates segmentation masks for an MS COCO image. im_id is
% the identifier for the image, and seg_idxs is an array of annotation
% indexes (according to the ordering in image_to_annotations) for that
% image. If seg_idxs is not given, all GT segmentations for the image are
% output. crowd is a logical value indicating whether crowd segments should
% be returned (the default value is false). masks is 3D logical array where 
% each plane is a GT segmentation, in the order given in seg_idxs. If 
% seg_idxs is not given, the indexing is according to the order in which 
% the GTs appear in image_to_annotations.
%
% Example:
%   cocoapi = CocoApi('instances_val2014.json');
%   masks = cocoGenGTMask(cocoapi, 544456, [1, 2, 4, 9, 10, 11]);
%   for idx = 1:size(masks,3)
%       imshow(masks(:,:,idx));
%       pause;
%   end
%
% @authors:     Ahmad Humayun
% @contact:     ahumayun@cc.gatech.edu
% @affiliation: Georgia Institute of Technology
% @date:        Spring 2015

    % get image info
    im_info = cocoapi.loadImgs(im_id);
    sz = [im_info.height, im_info.width];
    extra_info = struct;
    
%     im_filepath = fullfile(coco_data.image_folder, ...
%                            im_info.file_path, im_info.file_name);
%     I = imread(im_filepath);
%     imshow(I);
%     assert(size(I,1)==sz(1) && size(I,2)==sz(2), ...
%            'Something wrong with the image size');

    if ~exist('crowd', 'var')
        crowd = false;
    end
    
    % set to empty so getAnnIds gets both crowd and non-crowd annotations
    if crowd
        crowd = [];
    end
    
    % get the unique annotations for this image
    ann_ids = cocoapi.getAnnIds('imgIds',im_id, 'iscrowd',crowd);
    ann_ids = unique(ann_ids);
    
    % if segmentation idx not given, return all GTs in this image
    if ~exist('seg_idxs', 'var')
        num_gts = numel(ann_ids);
        seg_idxs = 1:num_gts;
    end
    
    % init a set of GT masks
    masks = false([sz, numel(seg_idxs)]);

    % get the segmentation for current GT
    anns = cocoapi.loadAnns(ann_ids);

    % iterate over all segmentations, and create GT for each
    for si = 1:numel(seg_idxs)
        % get current annotation
        ann = anns(seg_idxs(si));
        
        % get the stored bounding-box
        bb_coco = ann.bbox;
        bb_coco([1 2]) = bb_coco([1 2]) + 1;
        bb_coco = [bb_coco([2 1]), bb_coco([4 3])+bb_coco([2 1])-1];
        
        segs = ann.segmentation;
        % store the category id for the current GT
        extra_info(si).gt_category_id = ann.category_id;
        extra_info(si).bb_range = [inf inf -inf -inf];
        
        if ann.iscrowd
            masks(:,:,si) = cocoapi.decodeMask(segs);
%             bb_calc = bbFromMask(masks(:,:,aidx), bb_coco);
        else
            % iterate over each component of the current GT segmentation, 
            % and OR it with the segmentation mask, producing the final GT
%             for idx = 1:numel(segs)
%                 [masks(:,:,si), extra_info(si).bb_range] = ...
%                     polymask(segs{idx}, masks(:,:,si), extra_info(si).bb_range);
%             end
            
            masks(:,:,si) = cocoapi.segToMask(segs, sz(1), sz(2));
%             bb_calc = bbFromMask(masks(:,:,si), bb_coco);
        end
        
        % use the stored bounding box
        extra_info(si).bb_range = bb_coco;
        
%         if (strcmp(im_id, 'i578861') == 1 && si == 1) || ...
%                 (strcmp(im_id, 'i78060') == 1 && si == 5) || ...
%                 (strcmp(im_id, 'i72582') == 1 && si == 4) || ...
%                 (strcmp(im_id, 'i315173') == 1 && si == 3) || ...
%                 (strcmp(im_id, 'i532540') == 1 && si == 2) || ...
%                 (strcmp(im_id, 'i55594') == 1 && si == 16) || ...
%                 (strcmp(im_id, 'i134199') == 1 && si == 7) || ...
%                 (strcmp(im_id, 'i200365') == 1 && si == 3)
%             continue;
%         end
%         [r,c] = find(masks(:,:,si));
%         bb_calc = [min(r), min(c), max(r), max(c)];
%         assert(all(bb_calc == extra_info(si).bb_range));
%         % tl row, tl col, h, w
%         bb_calc([3 4]) = bb_calc([3 4]) - bb_calc([1 2]) + 1;
%         % tl col, tl row, w, h
%         bb_coco = coco_data.image_to_annotations.(im_id){seg_idx}.bbox;
%         bb_coco([1 2]) = bb_coco([1 2]) + 1;
%         if isempty(bb_calc)
%             bb_calc = [bb_coco([2 1]), 0, 0];
%         end
%         assert(all(abs(bb_calc - bb_coco([2 1 4 3])) <= 7), ...
%                'Some bb not right');
    end
    
    % check if any of the GT masks are empty - if so discard them
    empty_masks = ~squeeze(any(any(masks)));
    if any(empty_masks)
        masks(:,:,empty_masks) = [];
        extra_info(empty_masks) = [];
    end
end

function [ mask, curr_bb ] = polymask(seg, mask, curr_bb)
    assert(isvector(seg));
    % first row of poly is c/x and second row is r/y
    poly = reshape(seg, 2, numel(seg)/2) + 0.5;
%     poly(1,:) = poly(1,:) + 1;
    % other options for computing masks from polygons
%     curr_mask = poly2mask(poly(1,:), poly(2,:), ...
%                           size(mask,1), size(mask,2));
%     curr_mask = my_poly2mask(poly(1,:), poly(2,:), ...
%                              size(mask,1), size(mask,2));
    [curr_mask, bb] = poly2mask_mex(poly(1,:), poly(2,:), ...
                              size(mask,1), size(mask,2), 11);
    mask = mask | curr_mask;
    
    % adjust the bb
    curr_bb = [min([curr_bb([1 2]); bb([1 2])+1]), ...
               max([curr_bb([3 4]); bb([3 4])+1])];
    
% test code to check polygon shape and the mask produced
%     x = poly(1,:);  y = poly(2,:);
%     imshow(curr_mask);
%     hold on; patch(x, y, 'k', 'FaceColor','none', 'EdgeColor', 'r');
%     [c,r] = meshgrid(floor(min(x)):ceil(max(x)), floor(min(y)):ceil(max(y)));
%     hold on; plot(c(:), r(:), 'xb', 'MarkerSize',10);
end

function [mask] = my_poly2mask(x, y, h, w)
% creates a mask in the image by checking whether each pixel is inside the
% polygon defined by the given points or outside. A point that is inside
% would cross an odd number of line segments defining the polygon

    % if last point not same as first, then close contour
    if (x(end) ~= x(1)) || (y(end) ~= y(1))
        x(end+1) = x(1);
        y(end+1) = y(1);
    end
    
    yline_segs = [y(1:end-1)', y(2:end)'];
    xline_segs = [x(1:end-1)', x(2:end)'];
    
    s_xline_segs = sort(xline_segs,2);
    cond1 = bsxfun(@le, s_xline_segs(:,1), 1:w) & ...
            bsxfun(@gt, s_xline_segs(:,2), 1:w);
    
    tmp = bsxfun(@minus, 1:w, xline_segs(:,1));
    l = bsxfun(@rdivide, tmp, (xline_segs(:,2) - xline_segs(:,1)));
    
    yo = bsxfun(@times, l, yline_segs(:,2) - yline_segs(:,1));
    yt = bsxfun(@plus, yo, yline_segs(:,1));
    
    cond2a = (0 <= yt) & cond1;
    [z, c] = find(cond2a);
    cond2b = bsxfun(@le, yt(cond2a), 1:h);
    [tmp, r] = find(cond2b);
    % [r, c(tmp), z(tmp)] are the line segments crossed while travelling 
    % due north at each pixel
%     test = false([h,w,34]);
%     test(sub2ind([h,w,34], r, c(tmp), z(tmp))) = 1;
%     test = mod(sum(test,3), 2) == 1;
    
    linidxs = sub2ind([h, w], r, c(tmp));
    crossings = accumarray(linidxs, 1);
    crossings = [crossings; zeros(h*w - numel(crossings),1)];
    crossings = reshape(crossings, h, w);
    mask = mod(crossings,2) == 1;
    
%     assert(all(all(mask == test)));
end

function [bb] = bbFromMask(mask, stored_bb)
    stats = regionprops(mask, 'BoundingBox');
    if ~isempty(stats)
        bb = stats.BoundingBox;
        bb = [bb([2,1])+0.5, bb([2,1])+bb([4,3])-0.5];
    else
        % in case the mask had 0 pixels
        bb = stored_bb([1,2,1,2]);
    end
end