/* Copyright (C) 2013 Fuxin Li
 Fast version of segment overlap computation.
 *
 * Copyright (C) 2014 Ahmad Humayun
 * Changes:
 *   1. Added functionality to accept ignore regions for segms2
 *   2. Now can return bb overlap
 *   3. Consolidated overlap code into one function
 */

/*---
function [o12, obb] = fast_segm_overlap_mex(segms1, <segms2>, <ignore_rgs>)
Input:
	segms1 - binary 3D matrix, row*column*segments_1
	<segms2> - binary 3D matrix, row*column*segments_2. If not given, overlap
				is computed between segments of segm1.
    <ignore_rgs> - binary 3D matrix, row*column*segments_2
                    OR
                   binary 2D matrix, row*colum (when same ignore regions 
                                                for all segms2)
                * Omitting this is like sending an all zeros ignore_rgs, 
                *  i.e. nothing is ignored.
Output:
    o12 = segments_1 * segments_2 overlap matrix
    obb = segments_1 * segments_2 bb overlap matrix (ignore_rgs is not used to
    												 compute to the bounding
    												 box overlap)
--*/

#include "omp.h"
#include "mex.h"
#include "math.h"
#include "time.h"
#include "float.h"

typedef unsigned short ushort;
typedef unsigned long ulong;
typedef unsigned long long ptrincrtype;

void overlap_compute(const bool *segms, const bool *segms2, const bool* ignore_rgs, const size_t width, const size_t height, float *overlap, float *obb, 
					 const ulong& segm_sz, const ulong& bb_sz, const ushort& left, const ushort& top, const ushort& right, const ushort& bottom, 
					 const ulong& segm_sz2, const ulong& bb_sz2, const ushort& left2, const ushort& top2, const ushort& right2, const ushort& bottom2)
{
	ptrincrtype index_ik;
	ushort int_bbleft, int_bbtop, int_bbright, int_bbbottom;
	int res_int = 0;
	int ign_int = 0, ov_ign_int = 0;
// No overlap on bounding box means instant solution
	if (right < left2 || right2 < left || top > bottom2 || top2 > bottom)
	{
		*overlap = 0.0;
		if (obb)
			*obb = 0.0;
		return;
	}

// Kinds of pixels
// FPnocare + FPcare
// TPnocare + TPcare
// FNnocare + FNcare

// Now find the overlapping part of the bounding box and only compute res_int inside
// For left and top, we take the maximum, for right and bottom, we take the minimum
	int_bbleft = left > left2 ? left : left2;
	int_bbtop = top > top2 ? top : top2;
	int_bbright = right < right2 ? right : right2;
	int_bbbottom = bottom < bottom2 ? bottom : bottom2;
	for (int ii = int_bbleft; ii <= int_bbright; ++ii)
	{
		index_ik = ii * height;
        int this_res = 0;
        int this_ign = 0;
        int this_ign2 = 0;
#pragma omp parallel for reduction(+:this_res,this_ign,this_ign2)
		for (int jj = int_bbtop; jj <= int_bbbottom; ++jj)
		{
		// Only compute at the beginning (check if pixel ignored)
			if (segms[index_ik+jj] && segms2[index_ik+jj] && 
				(!ignore_rgs || !ignore_rgs[index_ik+jj]))
				++this_res;

			if (ignore_rgs && ignore_rgs[index_ik+jj]) {
				if (segms[index_ik+jj] && segms2[index_ik+jj]) {
					++this_ign2;
				} else if (segms[index_ik+jj] || segms2[index_ik+jj]) {
					++this_ign;
				}
			}	
		}
        res_int += this_res;
        ov_ign_int += this_ign2;
        ign_int += this_ign;
	}
	
	// find the number of ignored pixels in the non-overlapping region
	if (ignore_rgs) {
		// iterate over segms region (while avoid overlapping bb) and count number of ingnored pxls
		for (int ii = left; ii <= right; ++ii)
		{
			index_ik = ii * height;
	        int this_ign = 0;
	#pragma omp parallel for reduction(+:this_ign)
			for (int jj = top; jj <= bottom; ++jj)
			{
				// skip pixel if in the overlapping bb
				if (ii >= left2 && ii <= right2 && jj >= top2 && jj <= bottom2)
					continue;
				// if pixel ignored, and was in segm, add to the number of ignored pixels
				if (ignore_rgs[index_ik+jj])
					this_ign += segms[index_ik+jj];
			}
	        ign_int += this_ign;
		}
		// iterate over segms2 region (while avoid overlapping bb) and count number of ingnored pxls
		for (int ii = left2; ii <= right2; ++ii)
		{
			index_ik = ii * height;
	        int this_ign = 0;
	#pragma omp parallel for reduction(+:this_ign)
			for (int jj = top2; jj <= bottom2; ++jj)
			{
				// skip pixel if in the overlapping bb
				if (ii >= left && ii <= right && jj >= top && jj <= bottom)
					continue;
				// if pixel ignored, and was in segm2, add to the number of ignored pixels
				if (ignore_rgs[index_ik+jj])
					this_ign += segms2[index_ik+jj];
			}
	        ign_int += this_ign;
		}
	}

// Intersection / (A + B - A \int B)
	*overlap = (res_int - ov_ign_int) / (float(segm_sz + segm_sz2 - res_int - ov_ign_int - ign_int) + FLT_MIN);
// compute the bb overlap
	if (obb) {
		res_int = (int_bbright - int_bbleft + 1) * (int_bbbottom - int_bbtop + 1);
		*obb = res_int / (float(bb_sz + bb_sz2 - res_int) + FLT_MIN);
	}
}

void overlap_w_bb_filter(const bool *segms, const size_t width, const size_t height, const size_t nsegms, float *overlap, float *obb, 
						 const ulong *segm_sizes, const ulong *bb_sizes, const ushort *left, const ushort *top, const ushort *right, const ushort *bottom)
{
#pragma omp parallel for schedule(dynamic,2)
	for(int i=0; i<nsegms; ++i)
	{
		for(int j=i+1; j<nsegms; ++j)
		{
		    ptrincrtype index_ij = i * nsegms + j;
		    ptrincrtype index_ik = i * width * height;
		    ptrincrtype index_jk = j * width * height;
			overlap_compute((segms + index_ik), (segms + index_jk), NULL, width, height, 
							(overlap + index_ij), (obb ? obb + index_ij : obb), 
					 		segm_sizes[i], bb_sizes[i], left[i], top[i], right[i], bottom[i], 
					 		segm_sizes[j], bb_sizes[j], left[j], top[j], right[j], bottom[j]);
			ptrincrtype index_ji = j * nsegms + i;
			overlap[index_ji] = overlap[index_ij];
			// compute the bb overlap
			if (obb) {
				obb[index_ij] = obb[index_ji];
			}
		}
	}
	// set the diagonal to overlap of 1
	for (int i=0; i<nsegms; ++i) {
		overlap[i * nsegms + i] = 1.0;
		if (obb)
			obb[i * nsegms + i] = 1.0;
	}
}

void overlap_binary(const bool *segms, const bool *segms2, const bool* ignore_rgs, const bool& sngl_ignore, const size_t width, const size_t height, 
					const size_t nsegms, const size_t nsegms2, float *overlap, float *obb, 
					const ulong *segm_sizes, const ulong *bb_sizes, const ushort *left, const ushort *top, const ushort *right, const ushort *bottom, 
					const ulong *segm_sizes2, const ulong *bb_sizes2, const ushort *left2, const ushort *top2, const ushort *right2, const ushort *bottom2)
{
#pragma omp parallel for schedule(dynamic,2)
	for(int i=0; i<nsegms; ++i)
	{
		for(int j=0; j<nsegms2; ++j)
		{
		    ptrincrtype index_ij = j * nsegms + i;
		    ptrincrtype index_ik = i * width * height;
		    ptrincrtype index_jk = j * width * height;
		    const bool* curr_ignore_rgs = (sngl_ignore | !ignore_rgs) ? ignore_rgs : ignore_rgs + index_jk;
			overlap_compute((segms + index_ik), (segms2 + index_jk), curr_ignore_rgs, width, height, 
							(overlap + index_ij), (obb ? obb + index_ij : obb), 
					 		segm_sizes[i], bb_sizes[i], left[i], top[i], right[i], bottom[i], 
					 		segm_sizes2[j], bb_sizes2[j], left2[j], top2[j], right2[j], bottom2[j]);
		}
	}
}

/* Compute the bounding boxes for each segment */
/* Not fastest, test if fast enough */
void compute_bounding_box(const bool *segms, size_t width, size_t height, size_t nsegms,
        				  ulong *segm_size, ulong *bb_size, 
        				  ushort *left, ushort *top, ushort *right, ushort *bottom)
{
	int k;
    #pragma omp parallel for schedule(dynamic,2)
	for (k=0;k<nsegms;k++)
	{
		long cur_size = 0;
		short cur_left = width;
		short cur_top = height;
		short cur_right = 0;
        short cur_bottom = 0;
        size_t idx = width * height * k;
		for (int i=0;i<width;i++) {
			for (int j=0;j<height;j++)
			{
				if(segms[idx])
				{
					++cur_size;
					if (i < cur_left)
						cur_left = i;
					if (j < cur_top)
						cur_top = j;
					if (i > cur_right)
						cur_right = i;
					if (j > cur_bottom)
						cur_bottom = j;
				}
				++idx;
			}
		}
		segm_size[k] = cur_size;
		if (cur_size == 0) {
			left[k] = right[k] = top[k] = bottom[k] = 0;
			bb_size[k] = 0;
		} else {
			left[k] = cur_left, right[k] = cur_right, top[k] = cur_top, bottom[k] = cur_bottom;
			bb_size[k] = (cur_right - cur_left + 1) * (cur_bottom - cur_top + 1);
		}
	}
}

void mexFunction(int nargout, mxArray *out[], int nargin, const mxArray *in[]) 
{

    /* declare variables */
    mwSize width, height, nsegms, nsegms2;
    mwSize *sizes, *sizes2, *sizesign;
    bool *segms, *segms2, *ignore_rgs=NULL;
    float *o, *obb=NULL;
    
    /* check argument */
    if (nargin<1) {
        mexErrMsgTxt("One input argument required ( the segments in row * column * num_segments form )");
    }
    if (nargout>2) {
        mexErrMsgTxt("Too many output arguments");
    }
    if (!mxIsLogical(in[0]) || mxGetNumberOfDimensions(in[0]) < 2) {
        mexErrMsgTxt("Usage: segms must be a 3-dimensional logical matrix (if 2D, then it's considered h*w*1)");
    }
    /* sizes */
    sizes = (mwSize *)mxGetDimensions(in[0]);
    height = sizes[0];
    width = sizes[1];
	if (mxGetNumberOfDimensions(in[0]) == 3)
		nsegms = sizes[2];
	else
		nsegms = 1;
    segms = (bool *) mxGetData(in[0]);

	if (nargin > 1)
	{
		if (!mxIsLogical(in[1]) || mxGetNumberOfDimensions(in[1]) < 2) {
			mexErrMsgTxt("Usage: segms2 must be a 3-dimensional logical matrix (if 2D, then it's considered h*w*1)");
		}
		sizes2 = (mwSize *)mxGetDimensions(in[1]);
		if (sizes2[0] != height || sizes2[1] != width)
			mexErrMsgTxt("The two segment matrices must be from the same image (height and width must be the same)!");
		if (mxGetNumberOfDimensions(in[1]) == 3)
			nsegms2 = sizes2[2];
		else
			nsegms2 = 1;
		segms2 = (bool *) mxGetData(in[1]);
	}
    
    // this stores the indicator whether an ignore image is provided for each segms2 or just one for all
    bool sngl_ignore = false;
    if (nargin > 2 && !mxIsEmpty(in[2]))
    {
        // get the ignore regions arguments
        if (!mxIsLogical(in[2]) || mxGetNumberOfDimensions(in[2]) < 2) {
			mexErrMsgTxt("Usage: ignore_rgs must be a 3-dimensional logical matrix (if 2D, then it's considered h*w*1)");
		}
		sizesign = (mwSize *)mxGetDimensions(in[2]);
		if (sizesign[0] != height || sizesign[1] != width)
			mexErrMsgTxt("The ignore_rgs matrix must be from the same image (height and width must be the same)!");
        mwSize nignore;
		if (mxGetNumberOfDimensions(in[2]) == 3)
			nignore = sizesign[2];
		else {
			nignore = 1;
			sngl_ignore = true;
		}
        if (nignore > 1 && nignore != nsegms2)
            mexErrMsgTxt("The depth of ignore_rgs matrix should be the same as segms2 or 1!");
		ignore_rgs = (bool *) mxGetData(in[2]);
    }

	if (nargin > 1) {
		out[0] = mxCreateNumericMatrix(nsegms, nsegms2, mxSINGLE_CLASS, mxREAL);
        if (nargout > 1) {
            out[1] = mxCreateNumericMatrix(nsegms, nsegms2, mxSINGLE_CLASS, mxREAL);
        }
    } else {
		out[0] = mxCreateNumericMatrix(nsegms, nsegms, mxSINGLE_CLASS, mxREAL);
        if (nargout > 1) {
            out[1] = mxCreateNumericMatrix(nsegms, nsegms, mxSINGLE_CLASS, mxREAL);
        }
    }
    if (out[0]==NULL) {
	    mexErrMsgTxt("Not enough memory for the output matrix");
    }
    if (nargout > 1 && out[1]==NULL) {
	    mexErrMsgTxt("Not enough memory for the bb overlap output matrix");
    }

    o = (float *) mxGetPr(out[0]);
    if (nargout > 1)
    	obb = (float *) mxGetPr(out[1]);

	/* Not expecting an image has larger dimensions than 65536 * 65536 */
	ushort *bbleft = new ushort [nsegms];
	ushort *bbtop = new ushort [nsegms];
	ushort *bbright = new ushort [nsegms];
	ushort *bbbottom = new ushort [nsegms];
	ulong *segm_sizes = new ulong [nsegms];
	ulong *bb_sizes = new ulong [nsegms];
	compute_bounding_box(segms, width, height, nsegms, segm_sizes, bb_sizes, bbleft, bbtop, bbright, bbbottom);
	if (nargin > 1)
	{
		ushort *bbleft2 = new ushort [nsegms2];
		ushort *bbtop2 = new ushort [nsegms2];
		ushort *bbright2 = new ushort [nsegms2];
		ushort *bbbottom2 = new ushort [nsegms2];
		ulong *segm_sizes2 = new ulong [nsegms2];
		ulong *bb_sizes2 = new ulong [nsegms2];
		compute_bounding_box(segms2, width, height, nsegms2, segm_sizes2, bb_sizes2, bbleft2, bbtop2, bbright2, bbbottom2);
		overlap_binary(segms, segms2, ignore_rgs, sngl_ignore, width, height, nsegms, nsegms2, o, obb, 
					   segm_sizes, bb_sizes, bbleft, bbtop, bbright, bbbottom, 
					   segm_sizes2, bb_sizes2, bbleft2, bbtop2, bbright2, bbbottom2);
		delete[] bbleft2;
		delete[] bbright2;
		delete[] bbtop2;
		delete[] bbbottom2;
		delete[] segm_sizes2;
		delete[] bb_sizes2;
	}
	else
		overlap_w_bb_filter(segms, width, height, nsegms, o, obb, 
							segm_sizes, bb_sizes, bbleft, bbtop, bbright, bbbottom);
        
	delete[] bbleft;
	delete[] bbtop;
	delete[] bbright;
	delete[] bbbottom;
	delete[] segm_sizes;
	delete[] bb_sizes;
}
