/* This function generates a binary mask of polygon. The algorithm works by
projecting a ray north-bound from each pixel location, and then counting how
many times it intersects with the line segments of the polygon. If it
intersects an odd number of times, it is inside the polygon (and hence is 
marked as a 1 in the output), otherwise it is outside. To get better accuracy
for thin structures, the algorithm not only projects a ray from the center of
the pixel, rather it checks scale^2 number of sub-pixel locations centered at
the current pixel location.

The functionality of this algorithm is very similar to poly2mask in MATLAB's 
image processing toolbox. Although the algorithm itself is different (poly2mask
generates a sub-pixel accuracy polygon and checks if the center pixel passes
through this polygon).

@ args:
   xp: Vector of type double giving the x (column) coordinates of the polygon.
   yp: Vector of type double giving the y (row) coordinates of the polygon.
       Should be same size as xp.
     Note: if the last point is not a replica of the first point in xp,yp (for
           closure of the polygon), the code automatically replicates the first
           point.
   M: A positive scalar double (will be floored to the closest integer) giving
       the height of the output mask.
   N: A positive scalar double (will be floored to the closest integer) giving
       the width of the output mask.
   [scale]: (optional arg, default = 1) A positive odd parity scalar double, 
       indicating how much sub-pixels should be generated around the center 
       pixel. E.g. if scale = 5, the algorithm would search 25 locations around
       the center pixel, and declare success (that the center pixel is in the
       polygon) if rays from any of the 25 sub-pixel locations crosses an odd
       number of line segments defining the polygon.

 @output:
   mask: An MxN logical array, which is set to 1 at all pixel locations where 
       this algorithm deemed it was inside the polygon.
   bb_limits: Is a 1x4 int vector giving the minimum and maximum row,col where
       binary mask is set to 1. [min_r, min_c, max_r, max_c]. The indexing 
       starts from 0.

 Compilation:
   mex -O poly2mask_mex.cpp CFLAGS="\$CFLAGS -fopenmp" ...
    CXXFLAGS="\$CXXFLAGS -fopenmp" LDFLAGS="\$LDFLAGS -fopenmp" ...
    COMPFLAGS="\$COMPFLAGS /openmp";


 @authors:     Ahmad Humayun
 @contact:     ahumayun@cc.gatech.edu
 @affiliation: Georgia Institute of Technology
 @date:        Fall 2014
*/

#include <vector>
#include <limits>
#include <string.h>
#include <math.h>

#include "omp.h"
#include "mex.h"
#include "matrix.h"
#include "mexutils.h"

typedef unsigned int uint;
typedef std::vector<double> dblvec;

struct CommonData {
	uint scale;
	double max_d;
	double scale_inv;
	double* xpypdiffprod;
	double* sort_x1;
	double* sort_x2;

	CommonData(const dblvec& xp, const dblvec& yp, const uint& scale_) {
		scale = scale_;
		max_d = (double)(scale - 1) / (2 * scale);
		scale_inv = (double)1 / scale;
		xpypdiffprod = new double[xp.size()-1];
		sort_x1 = new double[xp.size()-1];
		sort_x2 = new double[xp.size()-1];
		for (size_t i = 0; i < xp.size()-1; ++i) {
			xpypdiffprod[i] = (double)(yp[i+1] - yp[i]) / (xp[i+1] - xp[i]);
			if (xp[i] < xp[i+1]) {
				sort_x1[i] = xp[i];
				sort_x2[i] = xp[i+1];
			} else {
				sort_x1[i] = xp[i+1];
				sort_x2[i] = xp[i];
			}
		}
	}

	~CommonData() {
		delete[] xpypdiffprod;
		delete[] sort_x1;
		delete[] sort_x2;
	}
};

inline double roundDbl(const double& num)
{
    return num < 0.0 ? ceil(num - 0.5) : floor(num + 0.5);
}

size_t checkDblVec(const mxArray* const arr, const char* errmsg)
{
	if (!mxIsDouble(arr) || !mexIsVector(arr)) {
    	mexErrMsgTxt(errmsg);
	}
	return mxGetNumberOfElements(arr);
}

double checkDblPositiveScalar(const mxArray* const arr, const char* errmsg)
{
	if (!mxIsDouble(arr) || !mexIsRealScalar(arr) || mxGetPr(arr)[0] <= 0) {
    	mexErrMsgTxt(errmsg);
	}
	return mxGetPr(arr)[0];
}

bool checkIfInPolygon(const size_t& xc, const size_t& yr, 
					  const dblvec& xp, const dblvec& yp,
					  const CommonData& cd) {
	// shoot a ray north-bound from (xc,yr) and check if it intersects odd 
	// number of line segments defining the polygon

	double x1, x2, y1, y2;
	double yt, r;

	uint* n_encounters = new uint[cd.scale];
	// std::vector<uint> n_encounters;

	for (double c = xc-cd.max_d; c <= xc+cd.max_d; c += cd.scale_inv) {
		memset(n_encounters, 0, sizeof(uint)*cd.scale);

		// iterate over all line segments
		for (size_t p = 0; p < xp.size() - 1; ++p) {
			// get coordinates of the current polygon border line segment
			x1 = xp[p];
			y1 = yp[p];
			x2 = xp[p+1];
			y2 = yp[p+1];

			// first check if this sub-pixel location is within the bounds of 
			// the line segment's x range
			if (cd.sort_x1[p] <= c && c < cd.sort_x2[p]) {
				yt = y1 + (c - x1) * cd.xpypdiffprod[p];

				// if intersection y, is not in the range [0,yr+max_d], check the
				// next polygon line segment
				if (yt < -2 || yt > yr+cd.max_d) {
					continue;
				}
				// in case yt is between 0 and yr-max_d, then add 1 to all
				if (yt <= yr-cd.max_d) {
					for (size_t idx = 0; idx < cd.scale; ++idx)
						++n_encounters[idx];
					continue;
				}
				// in case yt lies anywhere between yr-max_d and yr+max_d
				r = yr+cd.max_d;
				for (size_t idx = cd.scale-1; idx >= 0; --idx) {
					if (yt <= r)
						++n_encounters[idx];
					else
						break;
					r -= cd.scale_inv;
				}
			}
		}

		// now that we have collected number of crossings for each of the scaled
		// row indexes in this column, check if any if them is an odd number, and if 
		// so, this location is in the mask
		for (size_t idx = 0; idx < cd.scale; ++idx) {
			if (n_encounters[idx] % 2 == 1) {
				delete[] n_encounters;
				return true;
			}
		}
	}

	delete[] n_encounters;
	return false;
}

void adjustZeroSizePoly(const dblvec& xp, const dblvec& yp, 
						const size_t& M, const size_t& N, bool* out,
						int& min_r, int& max_r, int& min_c, int& max_c) {	
	// initialize limits
	max_r = max_c = std::numeric_limits<int>::min();
	min_r = min_c = std::numeric_limits<int>::max();

	// if no pixels were set in the mask, then iterate over all end-points of 
	// the polygon, marking all linearly interpolated points in-between
	for (size_t p = 0; p < xp.size() - 1; ++p) {
		double x1 = roundDbl(xp[p] - 1);
		double y1 = roundDbl(yp[p] - 1);
		double x2 = roundDbl(xp[p+1] - 1);
		double y2 = roundDbl(yp[p+1] - 1);
		int num_ps, num_ys;
		double incr_x, incr_y;

		// find the range of pixels we would need to mark
		num_ps = x1 - x2;
		if (num_ps < 0) {
			num_ps = -num_ps;
		}
		num_ys = y1 - y2;
		if (num_ys < 0) {
			num_ys = -num_ys;
		}
		if (num_ys > num_ps)
			num_ps = num_ys;
		++num_ps;
		incr_x = (x2-x1) / (num_ps-1);
		incr_y = (y2-y1) / (num_ps-1);

		// now iterate over all pixels between the two current end-points, and
		// mark the mask inside the range
		double curr_x = x1;
		double curr_y = y1;
		for (int i = 0; i < num_ps; ++i) {
			// if inside the bounds of the image
			if (curr_x >= 0 && curr_x < N && curr_y >= 0 && curr_y < M) {
				int x = roundDbl(curr_x);
				int y = roundDbl(curr_y);
				out[x*M + y] = true;

				// set the bb limits
				if (y < min_r) min_r = y;
				if (y > max_r) max_r = y;
				if (x < min_c) min_c = x;
				if (x > max_c) max_c = x;
			}
			// increment the position
			curr_x += incr_x;
			curr_y += incr_y;
		}
	}
}

void mexFunction(int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[])
{
	size_t num_pnts, temp, M, N;
	int min_r, max_r, min_c, max_c;
	uint scale;
	bool marked;
	// Check number of arguments
    if (nrhs < 4 || nrhs > 5) {
        mexErrMsgTxt("This function requires 4 or 5 arguments");
    }
    if (nlhs > 2) {
    	mexErrMsgTxt("This function only returns two values");
    }

	num_pnts = checkDblVec(prhs[0], 
		"first argument should be a double vector indicating x/c coordinates of the polygon");
	temp = checkDblVec(prhs[1], 
		"second argument should be a double vector indicating y/r coordinates of the polygon");
	if (num_pnts != temp)
		mexErrMsgTxt("The number of points in x and y vectors should be equal");

	// get the x,y vectors
	double* x = mxGetPr(prhs[0]);
	double* y = mxGetPr(prhs[1]);

	M = (size_t)checkDblPositiveScalar(prhs[2], 
		"third argument should be a positive scalar double indicating the height of the mask");
	N = (size_t)checkDblPositiveScalar(prhs[3], 
		"fourth argument should be a positive scalar double indicating the width of the mask");
	if (nrhs == 5) {
		scale = (unsigned int)checkDblPositiveScalar(prhs[4], 
			"fifth argument should be a positive scalar double indicating the scale");
		if (scale % 2 == 0)
			mexErrMsgTxt("fifth argument (scale) should be an odd number");
	} else {
		scale = 1;
	}

	dblvec xp(x, x + num_pnts);
	dblvec yp(y, y + num_pnts);
 	// if last point not same as first, then close contour
	if (x[0] != x[num_pnts-1] || y[0] != y[num_pnts-1]) {
		xp.push_back(x[0]);
		yp.push_back(y[0]);
	}
	// precompute some values, which will be needed by checkIfInPolygon
	CommonData cd(xp, yp, scale);

	// double max_d = (double)(scale - 1) / (2 * scale);	
	// mexPrintf("x: %f,%f\t y %f,%f\t %dx%d\t %d (%f) - closed contour %x\n", x[0], 
	// 		  x[num_pnts-1], y[0], y[num_pnts-1], M, N, scale, max_d, closed_cntr);

	// create the output array
	plhs[0] = mxCreateLogicalMatrix(M, N);
	bool* out = (bool*)mxGetData(plhs[0]);

	// initialize limits
	max_r = max_c = std::numeric_limits<int>::min();
	min_r = min_c = std::numeric_limits<int>::max();

	// iterate over the whole mask to see which pixels are inside the polygon
#pragma omp parallel for schedule(dynamic,2) reduction(min:min_r,min_c) reduction(max:max_r,max_c) private(marked)
	for (int c = 0; c < N; ++c) {
		marked = false;
		for (int r = 0; r < M; ++r) {
			out[c*M + r] = checkIfInPolygon(c+1, r+1, xp, yp, cd);
			// set marked if mask was set to true here
			marked |= out[c*M + r];
			// set the min/max for the row range
			if (out[c*M + r]) {
				if (r < min_r) min_r = r;
				if (r > max_r) max_r = r;
			}
		}
		// set the min/max for the column range
		if (marked) {
			if (c < min_c) min_c = c;
			if (c > max_c) max_c = c;
		}
	}

	// incase no pixel was set for the mask from the polygon
	if (min_r == std::numeric_limits<int>::max()) {
		adjustZeroSizePoly(xp, yp, M, N, out, min_r, max_r, min_c, max_c);
	}

	// if the user asked for the range, return it
	if (nlhs > 1) {
		// stores the limits of row/col
		plhs[1] = mxCreateNumericMatrix(1, 4, mxUINT32_CLASS, mxREAL);
		uint* range = (uint*)mxGetData(plhs[1]);
		range[0] = min_r;
		range[1] = min_c;
		range[2] = max_r;
		range[3] = max_c;
	}
}